#ifndef NDARRAY_CORE
#define NDARRAY_CORE

#include <librapid/config.hpp>

#include <librapid/errors.hpp>
#include <librapid/math/core_math.hpp>
#include <librapid/ndarray/basic_stride.hpp>
#include <librapid/ndarray/basic_extent.hpp>
#include <librapid/ndarray/to_string.hpp>
#include <librapid/ndarray/utils.hpp>
#include <librapid/ndarray/broadcast_arrays.hpp>
#include <librapid/ndarray/array_array_arithmetic.hpp>
#include <librapid/ndarray/array_scalar_arithmetic.hpp>
#include <librapid/ndarray/scalar_array_arithmetic.hpp>
#include <librapid/ndarray/array_arithmetic.hpp>
#include <librapid/ndarray/basic_ndarray.hpp>

#endif // NDARRAY_CORE